"""
Tests for Generator
"""
import io
import os
import shutil
import sys
import tempfile
from unittest import TestCase, mock

import pytest
from django.conf import settings

from dr_scaffold.generators import (
    Generator,
    AppGenerator,
    ModelGenerator,
    AdminGenerator,
    ViewGenerator,
    SerializerGenerator,
    FactoryGenerator,
    FactoryTestGenerator,
    URLGenerator,
    pluralize,
)
from dr_scaffold.scaffold_templates import (
    model_templates,
    serializer_templates,
)


class TestGenerator(TestCase):
    """
    Tests for Generator
    """

    test_settings = settings
    tmpdirpath = tempfile.mkdtemp()
    # tmpdirpath = "generated_tests_folder"
    core_folder = f"{tmpdirpath}/{test_settings.CORE_FOLDER}"
    api_folder = f"{tmpdirpath}/{test_settings.API_FOLDER}"

    def setUp(self):
        """
        Tests set up : get executed before each test
        """
        os.makedirs(self.tmpdirpath, exist_ok=True)
        os.makedirs(f"{self.core_folder}blog", exist_ok=True)
        os.makedirs(f"{self.api_folder}blog", exist_ok=True)
        for file_name in [
            f"{self.core_folder}blog/models.py",
            f"{self.api_folder}blog/serializers.py",
            f"{self.core_folder}blog/admin.py",
            f"{self.api_folder}blog/urls.py",
            f"{self.api_folder}blog/views.py",
        ]:
            with open(file_name, "x", encoding="utf8"):
                pass

    def tearDown(self):
        """
        Tests tear down : get executed after each test
        """
        for file_name in [
            f"{self.core_folder}blog/models.py",
            f"{self.api_folder}blog/serializers.py",
            f"{self.core_folder}blog/admin.py",
            f"{self.api_folder}blog/urls.py",
            f"{self.api_folder}blog/views.py",
        ]:
            os.remove(file_name)
        if os.path.exists(self.tmpdirpath):
            shutil.rmtree(self.tmpdirpath)

    @classmethod
    def test_pluralize(cls):
        """
        Test pluralization
        """
        assert pluralize("article") == "articles"
        assert pluralize("category") == "categories"
        assert pluralize("post") == "posts"

    @classmethod
    def test_init(cls):
        """
        Test arguments
        """
        generator_obj = Generator("blog", "Article", "", False, False)
        assert generator_obj.app_name == "blog"
        assert generator_obj.model_name == "Article"

    def test_add_setup_imports(self):
        """
        Tests imports
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, False
        )
        generator_obj.core_folder = self.core_folder
        files = (f"{self.core_folder}blog/models.py",)
        generator_obj.get_files = lambda: files
        generator_obj.get_file_imports = lambda: (serializer_templates.SETUP,)
        generator_obj.add_setup_imports()
        with open(files[0], "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body == serializer_templates.SETUP

    def test_setup_files(self):
        """
        Tests files setup
        """
        generator_obj = AppGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield", "body:textfield"),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        with mock.patch.object(
            AppGenerator, "is_already_generated", return_value=False
        ):
            generator_obj.generate_app()
        core_files = list(os.listdir(f"{self.core_folder}blog/"))
        api_files = list(os.listdir(f"{self.api_folder}blog/"))
        with open(f"{self.core_folder}blog/models.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert len(core_files + api_files) == 7  # taking in count migrations folder
        assert ("from django.db import models" in body) is True

    @classmethod
    @mock.patch("dr_scaffold.generators.file_api.is_present_in_file")
    def test_get_fields_string(cls, mock_is_in_file):
        """
        Tests
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, False
        )
        fields_string = generator_obj.get_fields_string()
        body_template = model_templates.TEXTFIELD % dict(name="body")
        title_template = model_templates.CHARFIELD % dict(name="title")
        string = title_template + body_template
        assert fields_string == string
        # test relation field type
        generator_obj2 = Generator(
            "blog", "Article", ("author:foreignkey:Author",), False, False
        )
        generator_obj2.get_fields_string()
        mock_is_in_file.assert_called_once()

    def test_get_fields_string_relation_model_not_created_yet(self):
        """
        Tests
        """
        generator_obj2 = Generator(
            "blog", "Article", ("author:foreignkey:Author",), False, False
        )
        generator_obj2.core_dir = self.core_folder
        generator_obj2.api_dir = self.api_folder
        captured_output = io.StringIO()
        sys.stdout = captured_output
        generator_obj2.get_fields_string()
        sys.stdout = sys.__stdout__
        assert (
            "⚠️ \x1b[93mbare in mind\x1b[0m that Author" in captured_output.getvalue()
        )

    @classmethod
    def test_get_model_string(cls):
        """
        Tests
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, False
        )
        fields_string = generator_obj.get_fields_string()
        model_string = generator_obj.get_model_string()
        string = model_templates.MODEL % (
            "Article",
            fields_string,
            "{self.title} {self.body}",
            "Articles",
        )
        assert model_string == string

    def test_run(self):
        """
        Tests
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, False
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        result = generator_obj.run()
        assert result == print("🎉 Your RESTful Article api resource is ready 🎉")
        # test case where something is wrong
        generator_obj2 = Generator("blog", "Author", ("title:charfiel",), False, False)
        generator_obj2.core_dir = self.core_folder
        generator_obj2.api_dir = self.api_folder
        result = generator_obj2.run()
        assert result == print("🤔 Oops something is wrong: charfiel")

    @mock.patch("dr_scaffold.generators.Generator.generate_app")
    def test_generate_api(self, mock_generate_app):
        """
        Tests
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, False
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate()
        mock_generate_app.assert_called_once()

    def test_generate_models(self):
        """
        Tests
        """
        generator_obj = ModelGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield", "body:textfield"),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_models()
        with open(f"{self.core_folder}blog/models.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert ("Article" in body) is True
        assert ("Articles" in body) is True
        assert ("title" in body) is True
        assert ("body" in body) is True
        # test when model already generated
        generator_obj = ModelGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield", "body:textfield"),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_models()
        assert body.count("class Article") == 1
        assert body.count("class Meta:") == 1

    def test_get_admin_parts(self):
        """
        Tests
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield",), ["C", "R"], False
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        head, body = generator_obj.get_admin_parts()
        assert ("import Article" in head) is True
        assert ("@admin.register(Article)" in body) is True

    def test_register_models_to_admin(self):
        """
        Tests
        """
        generator_obj = AdminGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_admin()
        with open(f"{self.core_folder}blog/admin.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert ("import Article" in body) is True
        assert ("@admin.register(Article)" in body) is True
        # test when model already registered
        generator_obj = AdminGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_admin()
        with open(f"{self.core_folder}blog/admin.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body.count("models import Article") == 1
        assert body.count("@admin.register(Article)") == 1

    def test_get_viewset_parts(self):
        """
        Tests
        """
        generator_obj = Generator("blog", "Article", ("title:charfield",), False, False)
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        head, body = generator_obj.get_viewset_parts()
        assert ("import Article" in head) is True
        assert ("class ArticleViewSet" in body) is True

    def test_generate_views(self):
        """
        Tests
        """
        generator_obj = ViewGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_views()
        with open(f"{self.api_folder}blog/views.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert ("import Article" in body) is True
        assert ("ArticleViewSet" in body) is True
        assert ("ArticleSerializer" in body) is True
        # test if view already exists
        generator_obj = ViewGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_views()
        with open(f"{self.api_folder}blog/views.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body.count("models import Article") == 1
        assert body.count("ArticleViewSet") == 1
        assert body.count("import ArticleSerializer") == 1
        assert body.count("serializer_class = ArticleSerializer") == 1

    def test_generate_views_with_mixins(self):
        """
        Tests
        """
        generator_obj = ViewGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=["C", "R"],
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_views()
        with open(f"{self.api_folder}blog/views.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert ("import Article" in body) is True
        assert ("ArticleViewSet" in body) is True
        assert ("def create" in body) is True
        assert ("def retrieve" in body) is True
        assert ("def update" in body) is not True

    def test_get_serializer_parts(self):
        """
        Tests
        """
        generator_obj = SerializerGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        head, body = generator_obj.get_serializer_parts()
        assert ("import Article" in head) is True
        assert ("class ArticleSerializer" in body) is True

    def test_generate_serializer(self):
        """
        Tests
        """
        generator_obj = SerializerGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_serializers()
        with open(
            f"{self.api_folder}blog/serializers.py", "r+", encoding="utf8"
        ) as file:
            body = "".join(file.readlines())
        assert ("import Article" in body) is True
        assert ("ArticleSerializer" in body) is True
        # test if serializer already exists
        generator_obj = SerializerGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_serializers()
        with open(
            f"{self.api_folder}blog/serializers.py", "r+", encoding="utf8"
        ) as file:
            body = "".join(file.readlines())
        assert body.count("import Article") == 1
        assert body.count("ArticleSerializer") == 1

    def test_get_url_parts(self):
        """
        Tests
        """
        generator_obj = URLGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        head, body = generator_obj.get_url_parts()
        assert ("import ArticleViewSet" in head) is True
        assert (", ArticleViewSet)" in body) is True

    @mock.patch("dr_scaffold.generators.file_api.replace_file_chunk")
    def test_generate_urls(self, mock_replace_chunk):
        """
        Tests
        """
        generator_obj = URLGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_urls()
        with open(f"{self.api_folder}blog/urls.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert ("import ArticleViewSet" in body) is True
        assert (", ArticleViewSet)" in body) is True
        # test : if url have been added we won't add it again
        generator_obj = URLGenerator(
            app_name="blog",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_urls()
        with open(f"{self.api_folder}blog/urls.py", "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body.count("import ArticleViewSet") == 1
        assert body.count(", ArticleViewSet)") == 1
        # test : if when creating another resource we replace the url_patterns
        generator_obj = URLGenerator(
            app_name="blog",
            model_name="Author",
            fields=("name:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate_urls()
        mock_replace_chunk.assert_called()

    @mock.patch("dr_scaffold.generators.Generator.generate")
    def test_generate_called(self, mock_generate):
        """
        Tests
        """
        generator_obj = Generator("blog", "Article", ("title:charfield",), False, False)
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.run()
        mock_generate.assert_called()

    def test_generate_app(self):
        """
        Tests
        """
        generator_obj = AppGenerator(
            app_name="blog2",
            model_name="Article",
            fields=("title:charfield",),
            mixins=False,
            tests=False,
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        setup_files_backup = generator_obj.setup_files
        with mock.patch.object(
            AppGenerator, "is_already_generated", return_value=False
        ):
            with mock.patch(
                "dr_scaffold.generators.AppGenerator.setup_files"
            ) as mocked:
                mocked.side_effect = setup_files_backup
                generator_obj.generate_app()
                mocked.assert_called()

    @mock.patch("dr_scaffold.generators.Generator.setup_files")
    def test_generate_app_appdir_exist(self, mock_setup_files):
        """
        Tests
        """
        generator_obj = Generator("blog", "Author", ("name:charfield",), False, False)
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.generate()
        mock_setup_files.assert_not_called()

    def test_get_folder_settings_not_set(self):
        """
        if no settings folder paths exists
        """
        delattr(self.test_settings, "CORE_FOLDER")
        delattr(self.test_settings, "API_FOLDER")
        generator = Generator("blog", "Author", ("name:charfield",), False, False)
        assert generator.core_dir == ""
        assert generator.api_dir == ""

    def test_get_folder_settings_without_forward_slash(self):
        """
        test add forward slash if forgotten
        """
        # overwriting settings for test
        setattr(self.test_settings, "CORE_FOLDER", "core")
        setattr(self.test_settings, "API_FOLDER", "api")
        with pytest.raises(ValueError, match=r".* should end with .*"):
            Generator("blog", "Author", ("name:charfield",), False, False)
        # settings to normal
        setattr(self.test_settings, "CORE_FOLDER", "my_app_core/")
        setattr(self.test_settings, "API_FOLDER", "my_app_api/")

    def test_get_folder_settings_with_forward_slash(self):
        """
        if settings are well set
        """
        setattr(self.test_settings, "CORE_FOLDER", "core/")
        setattr(self.test_settings, "API_FOLDER", "api/")
        generator_obj = Generator("blog", "Author", ("name:charfield",), False, False)
        assert generator_obj.core_dir == "core/"
        assert generator_obj.api_dir == "api/"

    def test_generate_factories(self):
        """
        Tests
        """
        generator_obj = FactoryGenerator(
            "blog", "Author", ("name:charfield",), False, True
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.root_dir = f"{self.tmpdirpath}"
        os.makedirs(f"{self.tmpdirpath}/tests/core/blog", exist_ok=True)
        with open(
            f"{self.tmpdirpath}/tests/core/blog/factories.py", "x", encoding="utf8"
        ):
            pass
        generator_obj.generate_factories()
        with open(
            f"{self.tmpdirpath}/tests/core/blog/factories.py", "r+", encoding="utf8"
        ) as file:
            body = "".join(file.readlines())
        assert body.count("AuthorFactory") == 1
        # test if factory already exists
        generator_obj = FactoryGenerator(
            "blog", "Author", ("name:charfield",), False, True
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.root_dir = f"{self.tmpdirpath}"
        generator_obj.generate_factories()
        with open(
            f"{self.tmpdirpath}/tests/core/blog/factories.py", "r+", encoding="utf8"
        ) as file:
            body = "".join(file.readlines())
        assert body.count("AuthorFactory") == 1

    def test_generate_factories_tests(self):
        """
        Tests
        """
        generator_obj = FactoryTestGenerator(
            "blog", "Author", ("name:charfield",), False, True
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.root_dir = f"{self.tmpdirpath}"
        os.makedirs(f"{self.tmpdirpath}/tests/core/blog", exist_ok=True)
        with open(
            f"{self.tmpdirpath}/tests/core/blog/test_factories.py", "x", encoding="utf8"
        ):
            pass
        generator_obj.generate_factories_tests()
        with open(
            f"{self.tmpdirpath}/tests/core/blog/test_factories.py",
            "r+",
            encoding="utf8",
        ) as file:
            body = "".join(file.readlines())
        assert body.count("def test_author_factory():") == 1
        # test if factory already exists
        generator_obj = FactoryTestGenerator(
            "blog", "Author", ("name:charfield",), False, True
        )
        generator_obj.core_dir = self.core_folder
        generator_obj.api_dir = self.api_folder
        generator_obj.root_dir = f"{self.tmpdirpath}"
        generator_obj.generate_factories_tests()
        with open(
            f"{self.tmpdirpath}/tests/core/blog/test_factories.py",
            "r+",
            encoding="utf8",
        ) as file:
            body = "".join(file.readlines())
        assert body.count("def test_author_factory():") == 1

    def test_add_setup_imports_with_tests(self):
        """
        Tests imports
        """
        generator_obj = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, False
        )
        assert len(generator_obj.get_files()) == 6
        assert len(generator_obj.get_file_imports()) == 6
        generator_obj2 = Generator(
            "blog", "Article", ("title:charfield", "body:textfield"), False, True
        )
        generator_obj2.core_dir = self.core_folder
        generator_obj2.api_dir = self.api_folder
        generator_obj2.root_dir = f"{self.tmpdirpath}"
        os.makedirs(f"{self.tmpdirpath}/tests/core", exist_ok=True)
        assert len(generator_obj2.get_files()) == 8
        assert len(generator_obj2.get_file_imports()) == 8
        generator_obj2.setup_folders()
        assert os.path.exists(f"{self.tmpdirpath}/tests/core/blog") is True



"""
templates for views
"""

VIEWSET = """class %(model)sViewSet(viewsets.ModelViewSet):
    queryset = %(model)s.objects.all()
    serializer_class = %(model)sSerializer

"""

MODEL_IMPORT = """from %(app)s.models import %(model)s
"""

SERIALIZER_IMPORT = """from %(app)s.serializers import %(model)sSerializer
"""

SETUP = """from rest_framework import mixins, permissions, viewsets
from rest_framework.response import Response

"""


CLRUD_MIXINS = {
    "C": """    mixins.CreateModelMixin,\n""",
    "R": """    mixins.ListModelMixin,\n    mixins.RetrieveModelMixin,\n""",
    "U": """    mixins.UpdateModelMixin,\n""",
    "D": """    mixins.DestroyModelMixin,\n""",
}

CLRUD_ACTIONS = {
    "C": """    def create(self, request, *args, **kwargs):
        serializer = %(model)sSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)\n\n""",
    "R": """    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        serializer = %(model)sSerializer(queryset, many=True)
        return Response(serializer.data)\n\n    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = %(model)sSerializer(instance=instance)
        return Response(serializer.data)\n\n""",
    "U": """    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = %(model)sSerializer(instance=instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)\n\n""",
    "D": """    def destroy(self, request, *args, **kwargs):
        return super().destroy(request=request, *args, **kwargs)\n\n""",
}

CLRUD_VIEWSET = """class %(model)sViewSet(
%(mixins)s\tviewsets.GenericViewSet
):
    queryset = %(model)s.objects.all()
    serializer_class = %(model)sSerializer
    #permission_classes = (permissions.IsAuthenticated,)

    def get_queryset(self):
        #user = self.request.user
        queryset = %(model)s.objects.all()
        #insert specific queryset logic here
        return queryset

    def get_object(self):
        #insert specific get_object logic here
        return super().get_object()

%(actions)s"""



"""
templates for models
"""

MODEL = """class %s(models.Model):\
    %s
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f"%s"

    class Meta:
        verbose_name_plural = "%s"

"""

MODEL_IMPORT = """from %(app)s.models import %(model)s
"""

CHARFIELD = """
    %(name)s = models.CharField(max_length=255, null=True, blank=True)"""

BIGINTEGERFIELD = """
    %(name)s = models.BigIntegerField()"""

BINARYFIELD = """
    %(name)s = models.BinaryField()"""

BOOLEANFIELD = """
    %(name)s = models.BooleanField()"""

DATEFIELD = """
    %(name)s = models.DateField()"""

FLOATFIELD = """
    %(name)s = models.FloatField()"""

GENERICIPADDRESSFIELD = """
    %(name)s = models.GenericIPAddressField()"""

NULLBOOLEANFIELD = """
    %(name)s = forms.NullBooleanField()"""

POSITIVEINTEGERFIELD = """
    %(name)s = models.PositiveIntegerField()"""

POSITIVESMALLINTEGERFIELD = """
    %(name)s = models.PositiveSmallIntegerField()"""

EMAILFIELD = """
    %(name)s = models.EmailField(max_length=254)"""

SLUGFIELD = """
    %(name)s = models.SlugField(max_length=200)"""

TIMEFIELD = """
    %(name)s = models.TimeField()"""

URLFIELD = """
    %(name)s = models.URLField(max_length=200)"""

SMALLINTEGERFIELD = """
    %(name)s = models.SmallIntegerField()"""

TEXTFIELD = """
    %(name)s = models.TextField(null=True, blank=True)"""

INTEGERFIELD = """
    %(name)s = models.IntegerField(null=True, default=0)"""

DECIMALFIELD = """
    %(name)s = models.DecimalField(max_digits=5, decimal_places=2, null=True, default=0.0)"""

DATETIMEFIELD = """
    %(name)s = models.DateTimeField(null=True)"""

FOREIGNKEY = """
    %(name)s = models.ForeignKey(%(related)s, on_delete=models.CASCADE, null=True)"""

MANYTOMANY = """
    %(name)s = models.ManyToManyField(%(related)s)"""

ONETOONE = """
    %(name)s = models.OneToOneField(%(related)s, on_delete = models.CASCADE, primary_key = True)"""

FIELD_TYPES = {
    "charfield": CHARFIELD,
    "textfield": TEXTFIELD,
    "integerfield": INTEGERFIELD,
    "decimalfield": DECIMALFIELD,
    "datetimefield": DATETIMEFIELD,
    "foreignkey": FOREIGNKEY,
    "bigintegerfield": BIGINTEGERFIELD,
    "binaryfield": BINARYFIELD,
    "booleanfield": BOOLEANFIELD,
    "datefield": DATEFIELD,
    "floatfield": FLOATFIELD,
    "genericipaddressfield": GENERICIPADDRESSFIELD,
    "nullbooleanfield": NULLBOOLEANFIELD,
    "positiveintegerfield": POSITIVEINTEGERFIELD,
    "positivesmallintegerfield": POSITIVESMALLINTEGERFIELD,
    "emailfield": EMAILFIELD,
    "slugfield": SLUGFIELD,
    "timefield": TIMEFIELD,
    "urlfield": URLFIELD,
    "smallintegerfield": SMALLINTEGERFIELD,
    "manytomany": MANYTOMANY,
    "onetoone": ONETOONE,
}

SETUP = """from django.db import models

"""



"""
templates for factories
"""

FACTORY = """class %(model)sFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = %(model)s

    def __new__(cls, *args, **kwargs):
        return super().__new__(*args, **kwargs)

"""

MODEL_IMPORT = """from %(app)s.models import %(model)s
"""

SETUP = """import factory

"""



"""
Tests for management command
"""
import os
import tempfile
from os import path
from unittest import TestCase

from dr_scaffold import file_api


class TestFileApi(TestCase):
    """
    Tests for file api
    """

    tmpfilepath = os.path.join(tempfile.gettempdir(), "tmp-testfile")

    def setUp(self):
        """
        Tests for file api
        """
        with open(self.tmpfilepath, "x", encoding="utf8") as file:
            file.write("Delete me!")

    def tearDown(self):
        """
        Tests for file api
        """
        os.remove(self.tmpfilepath)
        if path.exists("file.txt"):
            os.remove("file.txt")

    @classmethod
    def test_create_file(cls):
        """
        Tests for file api
        """
        file_api.create_file("file.txt")
        assert path.exists("file.txt")

    def test_wipe_file_content(self):
        """
        Tests for file api
        """
        file_api.wipe_file_content(self.tmpfilepath)
        self._extracted_from_test_wipe_files_4()

    def test_get_file_content(self):
        """
        Tests for file api
        """
        body = file_api.get_file_content(self.tmpfilepath)
        assert body == "Delete me!"

    def test_set_file_content(self):
        """
        Tests for file api
        """
        file_api.set_file_content(self.tmpfilepath, "new content")
        with open(self.tmpfilepath, "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body == "new content"

    def test_prepend_file_content(self):
        """
        Tests for file api
        """
        file_api.prepend_file_content(self.tmpfilepath, "new content")
        with open(self.tmpfilepath, "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body == "new contentDelete me!"

    def test_append_file_content(self):
        """
        Tests for file api
        """
        file_api.append_file_content(self.tmpfilepath, "new content")
        with open(self.tmpfilepath, "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body == "Delete me!new content"

    def test_wrap_file_content(self):
        """
        Tests for file api
        """
        file_api.wrap_file_content(self.tmpfilepath, "head", "tail")
        with open(self.tmpfilepath, "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body == "headDelete me!tail"

    def test_replace_file_chunk(self):
        """
        Tests for file api
        """
        file_api.replace_file_chunk(self.tmpfilepath, "Delete", "Remove")
        with open(self.tmpfilepath, "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body == "Remove me!"

    def test_is_present_in_file(self):
        """
        Tests for file api
        """
        assert file_api.is_present_in_file(self.tmpfilepath, "Delete") is True
        assert file_api.is_present_in_file(self.tmpfilepath, "Remove") is False

    def test_wipe_files(self):
        """
        Tests for file api
        """
        file_api.wipe_files((self.tmpfilepath,))
        self._extracted_from_test_wipe_files_4()

    def _extracted_from_test_wipe_files_4(self):
        with open(self.tmpfilepath, "r+", encoding="utf8") as file:
            body = "".join(file.readlines())
        assert body != "Delete me!"
        assert len(body) == 0

    @classmethod
    def test_create_files(cls):
        """
        Tests for file api
        """
        file_api.create_files(("file.txt",))
        assert path.exists("file.txt") is True



# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-ancestors
"""
Generator module where the real work happens
"""
from os import makedirs, path
from typing import Tuple

import inflect
import isort
from django.conf import settings

from dr_scaffold import file_api, print_off
from dr_scaffold.scaffold_templates import (
    admin_templates,
    app_template,
    factory_templates,
    factory_tests_templates,
    model_templates,
    serializer_templates,
    url_templates,
    view_templates,
)


def pluralize(string):
    """
    pluralizes a string word using a python library, needed for verbose model names and url paths
    """
    pluralizer = inflect.engine()
    return pluralizer.plural(string)


class BaseGenerator:
    """
    Base generator that any other generator class should inherit from
    """

    app_name: str
    model_name: str
    fields: Tuple[str]
    core_dir: str
    api_dir: str
    root_dir: str
    mixins: Tuple[str]
    tests: bool

    def __init__(self, app_name, model_name, fields, mixins, tests):
        self.init_paths_from_settings()
        self.app_name = app_name
        self.model_name = model_name
        self.fields = fields
        self.mixins = mixins
        self.tests = tests
        self.root_dir = "."

    @property
    def core_app_path(self):
        """
        Shortcut for core_dir + app_name
        """
        return self.core_dir + self.app_name

    @property
    def api_app_path(self):
        """
        Shortcut for api_dir + app_name
        """
        return self.api_dir + self.app_name

    @property
    def tests_path(self):
        """
        Shortcut for tests path
        """
        return self.root_dir + "/tests"

    def init_paths_from_settings(self):
        """
        Get folder paths from settings if they exist and add forward slash if forgotten
        """
        self.core_dir = getattr(settings, "CORE_FOLDER", "")
        self.api_dir = getattr(settings, "API_FOLDER", "")
        slashed = self.core_dir.endswith("/") and self.api_dir.endswith("/")
        if len(self.api_dir) > 0 and len(self.core_dir) > 0 and not slashed:
            raise ValueError("🤔 Oops CORE_FOLDER & API_FOLDER should end with a '/'")


class AppGenerator(BaseGenerator):
    """
    APP GENERATION
    1 - first we generate a django app through django's startapp command
    2 - we move the generated app directory to the apps directory if one is
     specified in our CLI command
    3 - we setup the files with the basic imports needed for each component
    4 - if application folder does already exist we return
    """

    def get_file_imports(self):
        """
        Returns all import statements
        """
        imports = [
            serializer_templates.SETUP,
            url_templates.SETUP,
            model_templates.SETUP,
            admin_templates.SETUP,
            view_templates.SETUP,
            app_template.TEMPLATE
            % (
                self.app_name.capitalize(),
                self.core_app_path.replace("/", "."),
            ),
        ]
        if self.tests:
            imports = [factory_templates.SETUP, factory_tests_templates.SETUP] + imports
        return imports

    def get_files(self):
        """
        Files to be generated
        """
        files = [
            f"{self.api_app_path}/serializers.py",
            f"{self.api_app_path}/urls.py",
            f"{self.core_app_path}/models.py",
            f"{self.core_app_path}/admin.py",
            f"{self.api_app_path}/views.py",
            f"{self.core_app_path}/apps.py",
        ]
        if self.tests:
            files = [
                f"{self.tests_path}/core/{self.app_name}/factories.py",
                f"{self.tests_path}/core/{self.app_name}/test_factories.py",
            ] + files
        return files

    def setup_folders(self):
        """
        creates folders if they not exist
        """
        migrations_path = f"{self.core_app_path}/migrations/"
        makedirs(self.api_app_path, exist_ok=True)
        makedirs(self.core_app_path, exist_ok=True)
        makedirs(migrations_path, exist_ok=True)
        if self.tests:
            factories_path = f"{self.tests_path}/core/{self.app_name}/"
            makedirs(factories_path, exist_ok=True)

    def add_setup_imports(self):
        """
        adds the bare minimum imports needed for each file
        """
        for i, file_path in enumerate(self.get_files()):
            file_api.set_file_content(file_path, self.get_file_imports()[i])

    def setup_files(self):
        """
        creates files if not exist, and adds appropriate imports
        """
        files = self.get_files()
        file_api.create_files(files)
        if not path.exists(f"{self.core_app_path}/migrations/__init__.py"):
            file_api.create_file(f"{self.core_app_path}/migrations/__init__.py")
        file_api.wipe_files(files)

    def is_already_generated(self):
        """
        Check if the core app path and the api app path exists already
        """
        return path.exists(self.core_app_path) or path.exists(self.api_app_path)

    def generate_app(self):
        """
        Generates files related to the app
        """
        if not self.is_already_generated():
            self.setup_folders()
            self.setup_files()
            self.add_setup_imports()


class ModelGenerator(BaseGenerator):
    """
    MODELS GENERATION METHODS
    1 - get_fields_string : match fields from cli with their templates and return a
     string of them joined
    2 - get_model_string : yield the fields string in the Model template and returns
     a new Model string
    3 - generate_models: check if the model does already exists in the models.py
     file if not it append it
    """

    def get_fields_string(self):
        """
        get appropriate fields templates based on the field type and return them joined in a string
        """
        actual_fields = []
        relation_types = ("foreignkey", "manytomany", "onetoone")
        file = f"{self.core_app_path}/models.py"
        for field in self.fields:
            field_name = field.split(":")[0]
            field_type = field.split(":")[1].lower()
            field_dict = dict(
                name=field_name,
                related=field.split(":")[2] if (field_type in relation_types) else "",
            )
            if field_type in relation_types:
                # tells developer if related model doesn't exist in file
                chunk = f'class {field_dict["related"]}(models.Model)'
                if not file_api.is_present_in_file(file, chunk):
                    print(
                        "⚠️ \033[93mbare in mind\033[0m that "
                        f"{field_dict['related']} model doesn't exist yet!"
                    )
            field_template = model_templates.FIELD_TYPES[field_type] % field_dict
            actual_fields.append(field_template)
        return "".join(actual_fields)

    def get_model_string(self):
        """
        returns a Model class string with fields and Meta class
        """
        fields_string = self.get_fields_string()
        fields_for_str = " ".join("{self.%s}" % f.split(":")[0] for f in self.fields)
        params = (
            self.model_name,
            fields_string,
            fields_for_str,
            pluralize(self.model_name.lower()).capitalize(),
        )
        return model_templates.MODEL % params

    def generate_models(self):
        """
        Generates models
        """
        file = f"{self.core_app_path}/models.py"
        chunk = f"class {self.model_name}"
        if file_api.is_present_in_file(file, chunk):
            return
        model_string = self.get_model_string()
        file_api.append_file_content(file, model_string)


class AdminGenerator(BaseGenerator):
    """
    MODELS REGISTRATION TO ADMIN
    1 - get_admin_parts : returns the register template and imports of the Model
     class
    2 - register_models_to_admin : check if the model already registered in admin.py
     if not it wraps the admin file content between the Model imports and register
     template
    """

    def get_admin_parts(self):
        """
        returns admin Model register template and import
        """
        app_dir = self.core_app_path
        app_path = app_dir.replace("/", ".")
        model_register_template = admin_templates.REGISTER % {"model": self.model_name}
        model_import_template = admin_templates.MODEL_IMPORT % {
            "app": app_path,
            "model": self.model_name,
        }
        return model_import_template, model_register_template

    def generate_admin(self):
        """
        Generate admin classes
        """
        file = f"{self.core_app_path}/admin.py"
        chunk = f"@admin.register({self.model_name})"
        if file_api.is_present_in_file(file, chunk):
            return
        head, body = self.get_admin_parts()
        file_api.wrap_file_content(file, head, body)


class SerializerGenerator(BaseGenerator):
    """
    SERIALIZER GENERATION
    1 - get_serializer_parts : returns the serializer template and imports of the Model class
    2 - generate_serializers : check if the viewset already exists in serializers.py if not
    it wraps the file content between the imports and the serializer template
    """

    def get_serializer_parts(self):
        """
        returns serializer class template and model import
        """
        app_dir = self.core_app_path
        app_path = app_dir.replace("/", ".")
        serializer_template = serializer_templates.SERIALIZER % {
            "model": self.model_name
        }
        imports = serializer_templates.MODEL_IMPORT % {
            "app": app_path,
            "model": self.model_name,
        }
        return imports, serializer_template

    def generate_serializers(self):
        """Generates serializers classes"""
        serializer_file = f"{self.api_app_path}/serializers.py"
        serializer_head = f"class {self.model_name}Serializer"
        if file_api.is_present_in_file(serializer_file, serializer_head):
            return
        head, body = self.get_serializer_parts()
        file_api.wrap_file_content(serializer_file, head, body)


class ViewGenerator(BaseGenerator):
    """
    VIEWS GENERATION
    1 - get_viewset_parts : returns the viewset template and model + serializers
    imports of the Model class
    2 - generate_views : check if the viewset already exists in views.py if not
    it wraps the file content between the imports and the viewset template
    """

    def get_mixins_template(self):
        """
        returns viewset template matching the mixins based on developer command
        """
        mixins_list = {key: view_templates.CLRUD_MIXINS[key] for key in self.mixins}
        mixins_string = "".join(
            str(mixins_list[x] % {"model": self.model_name}) for x in mixins_list
        )
        actions_list = {key: view_templates.CLRUD_ACTIONS[key] for key in self.mixins}
        actions_string = "".join(
            str(actions_list[x] % {"model": self.model_name}) for x in mixins_list
        )
        return view_templates.CLRUD_VIEWSET % {
            "model": self.model_name,
            "mixins": mixins_string,
            "actions": actions_string,
        }

    def get_viewset_parts(self):
        """
        returns viewsets templates and imports
        """
        core_app_path = self.core_app_path.replace("/", ".")
        api_app_path = self.api_app_path.replace("/", ".")
        viewset_template = view_templates.VIEWSET % {"model": self.model_name}
        if self.mixins:
            viewset_template = self.get_mixins_template()

        model_import_template = view_templates.MODEL_IMPORT % {
            "app": core_app_path,
            "model": self.model_name,
        }
        serializer_import_template = view_templates.SERIALIZER_IMPORT % {
            "app": api_app_path,
            "model": self.model_name,
        }
        imports = model_import_template + serializer_import_template
        return imports, viewset_template

    def generate_views(self):
        """
        Generates views classes
        """
        file = f"{self.api_app_path}/views.py"
        chunk = f"class {self.model_name}ViewSet"
        if file_api.is_present_in_file(file, chunk):
            return
        head, body = self.get_viewset_parts()
        file_api.wrap_file_content(file, head, body)


class FactoryGenerator(BaseGenerator):
    """
    FACTORY GENERATION
    1 - get_factory_parts : returns the factory template and imports of the Model class
    2 - generate_factories : check if the factory already exists in factories.py if not
    it wraps the file content between the imports and the factory template
    """

    def get_factory_parts(self):
        """
        returns serializer class template and model import
        """
        app_dir = self.core_app_path
        app_path = app_dir.replace("/", ".")
        factory_template = factory_templates.FACTORY % {"model": self.model_name}
        imports = factory_templates.MODEL_IMPORT % {
            "app": app_path,
            "model": self.model_name,
        }
        return imports, factory_template

    def generate_factories(self):
        """Generates serializers classes"""
        if not self.tests:
            return
        factory_file = f"{self.tests_path}/core/{self.app_name}/factories.py"
        factory_head = f"class {self.model_name}Factory"
        if file_api.is_present_in_file(factory_file, factory_head):
            return
        head, body = self.get_factory_parts()
        file_api.wrap_file_content(factory_file, head, body)


class FactoryTestGenerator(BaseGenerator):
    """
    FACTORY GENERATION
    1 - get_factory_parts : returns the factory template and imports of the Model class
    2 - generate_factories : check if the factory already exists in factories.py if not
    it wraps the file content between the imports and the factory template
    """

    def get_factory_test_parts(self):
        """
        returns serializer class template and model import
        """
        factory_test_template = factory_tests_templates.FACTORY_TEST % {
            "model_lower": self.model_name.lower(),
            "model": self.model_name,
        }
        imports = factory_tests_templates.MODEL_IMPORT % {
            "model": self.model_name,
        }
        return imports, factory_test_template

    def generate_factories_tests(self):
        """Generates serializers classes"""
        if not self.tests:
            return
        factory_test_file = f"{self.tests_path}/core/{self.app_name}/test_factories.py"
        factory_test_head = f"def test_{self.model_name.lower()}_factory():"
        if file_api.is_present_in_file(factory_test_file, factory_test_head):
            return
        head, body = self.get_factory_test_parts()
        file_api.wrap_file_content(factory_test_file, head, body)


class URLGenerator(BaseGenerator):
    """
    URLS GENERATION
    1 - get_url_parts : returns the url template and model imports of the Model class
    2 - generate_urls : check if the url already exists in urls.py if not
    it wraps the file content between the imports and the url template, before wrapping
    it checks if the file has URL_PATTERS it take it off and append it to the url template
    """

    def get_url_parts(self):
        """
        returns the url template and imports of the Model class
        """
        api_app_dir = self.api_app_path
        api_app_path = api_app_dir.replace("/", ".")
        plural_path = pluralize(self.model_name.lower())
        url_template = url_templates.URL % {
            "model": self.model_name,
            "path": plural_path,
        }
        imports = url_templates.VIEWSET_IMPORT % {
            "app": api_app_path,
            "model": self.model_name,
        }
        return imports, url_template

    def generate_urls(self):
        """
        Generates urls
        """
        file = f"{self.api_app_path}/urls.py"
        chunk = f"{self.model_name}ViewSet)"
        if file_api.is_present_in_file(file, chunk):
            return
        head, body = self.get_url_parts()
        if file_api.is_present_in_file(file, url_templates.URL_PATTERNS):
            file_api.replace_file_chunk(file, url_templates.URL_PATTERNS, "")
        body = body + url_templates.URL_PATTERNS
        file_api.wrap_file_content(file, head, body)


class Generator(
    AppGenerator,
    ModelGenerator,
    AdminGenerator,
    SerializerGenerator,
    ViewGenerator,
    FactoryGenerator,
    FactoryTestGenerator,
    URLGenerator,
):
    """
    A wrapper for CLI command arguments and the REST api different files generation methods
    """

    def __init__(self, app_name, model_name, fields, mixins, tests):
        super().__init__(
            app_name=app_name,
            model_name=model_name,
            fields=fields,
            mixins=mixins,
            tests=tests,
        )

    def run(self):
        """
        runs generate api and throws an exception if something went wrong
        """
        try:
            self.generate()
        except Exception as error:
            return print(f"🤔 \033[93mOops\033[0m something is wrong: {error}")
        return print(
            "🎉 \033[92mSuccess:\033[0m your RESTful "
            f"\033[1m{self.model_name}\033[0m api resource is ready."
        )

    def sort_imports(self):
        """
        Sort file imports using isort
        """
        print("🦄 \033[95mSorting file imports ...\033[0m")
        for file in self.get_files()[:-1]:
            with print_off.HiddenPrints():
                isort.file(file)

    def generate(self):
        """
        Calls the appropriate generator classes in the correct order
        """
        self.generate_app()
        self.generate_models()
        self.generate_admin()
        self.generate_serializers()
        self.generate_views()
        self.generate_factories()
        self.generate_factories_tests()
        self.generate_urls()
        self.sort_imports()



"""
Tests for management command
"""
from unittest import mock

import django
from django.core.management import call_command
from django.test import override_settings
from django.test.testcases import SimpleTestCase

django.setup()


@override_settings(
    INSTALLED_APPS=[
        "dr_scaffold",
    ],
)
class CommandTestCase(SimpleTestCase):
    """
    simple tests for management command
    """

    @classmethod
    @mock.patch("dr_scaffold.generators.Generator.run")
    def test_command(cls, mock_run):
        """
        Test for successful command call
        """
        call_command("dr_scaffold", "blog", "Article", "title:charfield", False)
        mock_run.assert_called()

    @classmethod
    @mock.patch("dr_scaffold.generators.Generator.generate")
    def test_generate_api(cls, mock_generate_api):
        """
        Test for successful command call
        """
        call_command("dr_scaffold", "blog", "Article", "title:charfield", False)
        mock_generate_api.assert_called()



"""
turns off prints
"""
import os
import sys


class HiddenPrints:
    """
    turns off prints
    """

    def __init__(self):
        """
        turns off prints
        """
        self._original_stdout = sys.stdout

    def __enter__(self):
        """
        turns off prints
        """
        sys.stdout = open(os.devnull, "w", encoding="utf8")

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        turns off prints
        """
        sys.stdout.close()
        sys.stdout = self._original_stdout



"""
templates for factories
"""

FACTORY_TEST = """@pytest.mark.django_db
def test_%(model_lower)s_factory():
    instance = %(model)sFactory()
    assert instance.id

"""

MODEL_IMPORT = """from factories import %(model)sFactory
"""

SETUP = """import pytest

"""
